/* 
 *    Programmed By: Mohammed Isam Mohammed [mohammed_isam1984@yahoo.com]
 *    Copyright 2014, 2015 (c)
 * 
 *    file: custom_func.c
 *    This file is part of tcalc.
 *
 *    tcalc is free software: you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation, either version 3 of the License, or
 *    (at your option) any later version.
 *
 *    tcalc is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with tcalc.  If not, see <http://www.gnu.org/licenses/>.
 */    

#include "defs.h"

char custom_func_name[MAX_FUNCTIONS][MAX_FUNCTION_NAME];
char *func_body[MAX_FUNCTIONS];
char *func_args[MAX_FUNCTIONS];
int func_args_count[MAX_FUNCTIONS];
int func_body_len[MAX_FUNCTIONS];
char var_name[MAX_VARIABLES][MAX_VARIABLE_LEN];
float var_value[MAX_VARIABLES];
int total_vars;
int total_custom_funcs = 0;

char global_var_name[MAX_VARIABLES][MAX_VARIABLE_LEN];
double global_var_value[MAX_VARIABLES];
int total_global_vars = 0;

char buf2[2048];

/*
 * Used for debugging.
 */
/*
void print_funcs()
{
  struct custom_func_s *f = &custom_func_head;
  int i = 1;
  while(f)
  {
    printf("[%d] 0x%x\tnext (0x%x)\tprev (0x%x)\n", i, f, f->next, f->prev);
    i++;
    f = f->next;
  }
}
*/

/*
 * Sadly we are using strtok() in input_file() function. And as
 * input_file() will most probably call process_custom_func() several
 * times during processing, using strtok() in process_custom_func()
 * is thus harmful, as it will cripple input_file(). So we use this
 * internal strtok()-like function for the purpose of process_custom_func()
 * only. It is not a mighty function like strtok(), but it works just enough.
 */
char *__my_strtok(char *str, char *delim)
{
  static char *my_str = (char *)NULL;
  if(str) my_str = str;
  char *res = my_str;
  char found = 0;
  while(*my_str)
  {
    int i;
    for(i = 0; i < strlen(delim); i++)
      if((*my_str) == delim[i])
      {
	(*my_str) = '\0';
	found = 1; break;
      }
    my_str++;
    if(found) break;
  }
  if(str == my_str) return (char *)NULL;
  if(found) return res;
  return (char *)NULL;
}

/*
 * Adds argument list to a function. Arguments are passed as a string,
 * which is broken down to its components and saved inside function
 * arguments.
 */
int add_function_args(struct custom_func_s *func, char *args, int len)
{
  func->func_args = strndup(args, len);
  if(!func->func_args)
  {
    return 0;
  }
  
  int i = func->total_vars;
  int argc = 0;
  
  int l, k;
  k = l = 0;
  while(is_whitespace(args[k])) k++;
  while(k < len)
  {
    while(args[k] != ',' && !is_whitespace(args[k]) && k < len) k++;
    if(k == 0) break;
    strncpy(func->var_name[i], args+l, k-l);
    func->var_name[i][k-l] = '\0';
    while(args[k] != ',' && k < len) k++;
    k++;
    while(is_whitespace(args[k])) k++;
    l = k;
    //printf("var %d = %s\n", i, func->var_name[i]);
    i++;
    argc++;
    if(i >= MAX_VARIABLES) return 0;
  }
  func->total_vars = i;
  func->func_args_count = argc;
  return 1;
}

/*
 * Deletes a function from the list.
 */
void delete_custom_func(struct custom_func_s *func)
{
  //printf("delete_custom_func():\n");
  //print_funcs();	/*******************************/
  if(!func) return;
  if(func->func_body) free(func->func_body);
  if(func->func_args) free(func->func_args);
  total_custom_funcs--;
  /* deleting first function in the list? */
  if(func == &custom_func_head)
  {
    /* last function in list? */
    if(func->next == NULL)
    {
      memset((void *)func, 0, sizeof(struct custom_func_s));
    }
    else
    {
      struct custom_func_s *n = func->next;
      memcpy((void *)func, (void *)func->next, sizeof(struct custom_func_s));
      func->prev = (struct custom_func_s *)NULL;
      func->next->prev = func;
      free(n);
    }
    //print_funcs();	/*******************************/
    return;
  }
  
  struct custom_func_s *prev = func->prev;
  struct custom_func_s *next = func->next;
  prev->next = next;
  if(next) next->prev = prev;
  free(func);
}

/*
 * Creates a new custom function and adds it to the list.
 */
struct custom_func_s *create_custom_func(char *name, int name_len, int argc)
{
  //printf("create_custom_func():\n");
  if(total_custom_funcs >= MAX_FUNCTIONS)
  {
    printf("Error: Maximum number of functions declared\n");
    return (struct custom_func_s *)NULL;
  }
  if(name_len >= MAX_FUNCTION_NAME)
  {
    printf("Error: Function name is too long\n");
    return (struct custom_func_s *)NULL;
  }
  //print_funcs();	/*******************************/
  struct custom_func_s *new_func;
  /* first function to be defined? */
  if(total_custom_funcs == 0)
  {
    new_func = &custom_func_head;
    memset((void *)new_func, 0, sizeof(struct custom_func_s));
    goto add;
  }
  new_func = (struct custom_func_s *)malloc(sizeof(struct custom_func_s));
  if(!new_func)
  {
    printf("Error: Insufficient memory to declare new function\n");
    printf("       Try 'undef func' to free some memory\n");
    return (struct custom_func_s *)NULL;
  }
  memset((void *)new_func, 0, sizeof(struct custom_func_s));
  struct custom_func_s *f = &custom_func_head;
  struct custom_func_s *old_f = f;
  while(f)
  {
    old_f = f;
    f = f->next;
  }
  old_f->next = new_func;
  new_func->prev = old_f;
add:
  strncpy(new_func->custom_func_name, name, name_len);
  new_func->custom_func_name[name_len] = '\0';
  new_func->func_args_count = argc;
  strcpy(new_func->var_name[0], new_func->custom_func_name);
  new_func->total_vars = 1;
  total_custom_funcs++;
  //print_funcs();	/*******************************/
  return new_func;
}

/*
 * Print a list of global or local variables, according to the passed
 * active_func argument.
 */
void print_vars(struct custom_func_s *active_func)
{
  int i;
  if(active_func == NULL)		/* global scope */
  {
    printf("Global variables:\n");
    if(!total_global_vars)
      printf("  (No global variables defined)\n");
    else
      for(i = 0; i < total_global_vars; i++)
	printf("  %s (%f)\n", global_var_name[i], global_var_value[i]);
  }
  else					/* function scope */
  {
    printf("Local variables (function %s()):\n", active_func->var_name[0]);
    if(active_func->total_vars == 1)
      printf("  (No variables defined)\n");
    else
      for(i = 1; i < active_func->total_vars; i++)
	printf("  %s (%f)\n", active_func->var_name[i], active_func->var_value[i]);
  }
}

/*
 * Determines if var is a local or a global variable or if
 * it is not defined all together.
 */
int is_var(char *var, int len, int *index, struct custom_func_s *active_func)
{
  if(!var || var[0] == '\n') return 0;
  int i;
  if(active_func == NULL) goto read_global;
  /* check local variables first */
  for(i = 0; i < active_func->total_vars; i++)
    if(len == strlen(active_func->var_name[i]) &&
       strncmp(var, active_func->var_name[i], len) == 0)
    {
      (*index) = i;
      return 1;
    }
read_global:
  /* try global variables */
  for(i = 0; i < total_global_vars; i++)
    if(len == strlen(global_var_name[i]) &&
       strncmp(var, global_var_name[i], len) == 0)
    {
      (*index) = i;
      return 2;
    }
  /* variable not found */
  return 0;
}

/*
 * Returns the value of var.
 */
double get_var(char *var, int len, struct custom_func_s *active_func, char *error)
{
  *error = 0;
  if(!var || var[0] == '\n') goto finish;
  int i;
  if(active_func == NULL) goto read_global;
  /* check local variables first */
  for(i = 0; i < active_func->total_vars; i++)
    if(len == strlen(active_func->var_name[i]) &&
       strncmp(var, active_func->var_name[i], len) == 0)
    {
      return active_func->var_value[i];
    }
read_global:
  /* try global variables */
  for(i = 0; i < total_global_vars; i++)
    if(len == strlen(global_var_name[i]) &&
       strncmp(var, global_var_name[i], len) == 0)
    {
      return global_var_value[i];
    }
finish:
  /* variable not found */
  *error = 1;
  return 0;
}

/*
 * Deletes var by giving its name.
 */
void delete_var_by_name(char *var, struct custom_func_s *active_func)
{
  if(!var || var[0] == '\n') return;
  int i;
  int len = strlen(var);
  if(active_func == NULL)		/* global scope */
  {
    for(i = 0; i < total_global_vars; i++)
      if(len == strlen(global_var_name[i]) &&
	 strcmp(var, global_var_name[i]) == 0)
      {
	delete_var(i, NULL);
	return;
      }
  }
  else					/* function scope */
  {
    for(i = 0; i < active_func->total_vars; i++)
      if(len == strlen(active_func->var_name[i]) &&
	 strcmp(var, active_func->var_name[i]) == 0)
      {
	delete_var(i, active_func);
	return;
      }
  }
  printf("Variable %s is not defined in the current scope\n", var);
}

/*
 * Deletes var by its index.
 */
void delete_var(int var_no, struct custom_func_s *active_func)
{
  int i;
  if(active_func == NULL)	/* global scope */
  {
    for(i = var_no; i < total_global_vars-1; i++)
    {
      strcpy(global_var_name[i], global_var_name[i+1]);
      global_var_value[i] = global_var_value[i+1];
    }
    total_global_vars--;
  }
  else				/* function scope */
  {
    for(i = var_no; i < active_func->total_vars-1; i++)
    {
      strcpy(active_func->var_name[i], active_func->var_name[i+1]);
      active_func->var_value[i] = active_func->var_value[i+1];
    }
    active_func->total_vars--;
  }
}


/*
 * Checks if func is a defined custom function.
 */
int is_custom_func(char *func, int len1)
{
  //int len1 = strlen(func);
  struct custom_func_s *f = &custom_func_head;
  //printf("[%s][%s]\n", func, f->custom_func_name);
  while(f)
  {
    int len2 = strlen(f->custom_func_name);
    if(len1 != len2)
    {
      f = f->next;
      continue;
    }
    if(strncmp(f->custom_func_name, func, len1) == 0)
    {
      printf("Error: Function %s() is already defined\n", f->custom_func_name);
      printf("Do you want to overwrite it? [y/N] ");
      char ans[10];
      char *s = fgets(ans, 4, stdin);
      if(s && (strcasecmp(ans, "yes\n") == 0 || strcasecmp(ans, "y\n") == 0))
      {
	delete_custom_func(f);
	return 0;
      }
      printf("Cancelled\n");
      return 1;
    }
    f = f->next;
  }
  return 0;
}


/*
 * Deletes a function given its name.
 */
void delete_custom_func_by_name(char *func_name)
{
  if(!func_name || func_name[0] == '\n') return;
  struct custom_func_s *f = &custom_func_head;
  while(f)
  {
    if(strcmp(func_name, f->custom_func_name) == 0)
    {
      delete_custom_func(f);
      return;
    }
    f = f->next;  
  }
  printf("Function %s() is not defined\n", func_name);
}

/*
 * Prints a list of defined custom functions.
 */
void print_custom_funcs()
{
  //int i;
  printf("Custom-defined functions:\n");
  if(!total_custom_funcs)
    printf("  (No functions defined)\n");
  else
  {
    struct custom_func_s *f = &custom_func_head;
    while(f)
    {
      if(f->func_args_count)
	 printf("  %s(%s)\n", f->custom_func_name, f->func_args);
      else
	 printf("  %s()\n", f->custom_func_name);
      f = f->next;
    }
  }
}

double read_var_value(char *tmp, char *error)
{
  *error = 0;
  double res = 0;
  if(tmp[0] == 'b' || tmp[0] == 'B')
    res = binary_to_decimal(tmp+1, error);
  else if((tmp[0] == '0' && tmp[1] == 'X') ||
          (tmp[0] == '0' && tmp[1] == 'x'))
    res = hex_to_decimal(tmp, error);
  else if(tmp[0] == '0'  && tmp[1] != '.')
    res = octal_to_decimal(tmp, error);
  else res = atof(tmp);
  return res;
}

/*
 * Does the donkey work of processing the function body
 * of a custom function, passed as *func.
 */
int process_custom_func(struct custom_func_s *func, char *args, double *res)
{
  char *str = strdup(func->func_body);
  int i;
  int k;
  char *tmp;
  char error = 0;
  
  /* extract passed arguments, if any */
  if(strlen(args))
  {
    if(func->func_args_count == 0)
    {
      printf("Error: Passing arguments to a parameterless function\n");
      free(str); return 0;
    }
    int l;
    k = l = 0;
    i = 1;
    char tmp_var[32];
    while(is_whitespace(args[k])) k++;

    while(k < strlen(args))
    {
      while(args[k] != ',' && !is_whitespace(args[k]) && k < strlen(args))
	k++;
      strncpy(tmp_var, args+l, k-l);
      tmp_var[k-l] = '\0';
      func->var_value[i] = read_var_value(tmp_var, &error);
      if(error) return 0;
      while(args[k] != ',' && k < strlen(args)) k++;
      if(k < strlen(args))
      {
	k++;
	while(is_whitespace(args[k])) k++;
	l = k;
      }
      i++;
    }
    i--;
    if(i != func->func_args_count)
    {
      printf("Error: Missing argument(s) to function %s()\n", func->custom_func_name);
      free(str); return 0;
    }
  }
  else
  {
    if(func->func_args_count)
    {
      printf("Error: Missing argument(s) to function %s()\n", func->custom_func_name);
      free(str); return 0;
    }
  }
  
  tmp = __my_strtok(str, "\n");
  if(!tmp)
  {
    printf("Error: Invalid function body for function %s()\n", func->custom_func_name);
    free(str); return 0;
  }

  SCOPE = FUNCTION_SCOPE(1);
  goto parse_line;
  
  while((tmp = __my_strtok(NULL, "\n")))
  {
parse_line:
    if(0) ;	//to shut up gcc
    char COMMENT = 0;
    tmp = trim(tmp, 0);
    i = 0;
    if(tmp[i] == '#') { COMMENT = 1; continue; }
    
    int p = proc_input(tmp, func);
    if(p > 0) break;		/* means there was an error */
    else if(!p) continue;	/* means everything was processed right */
    /* NOTE: negative return value from proc_input() means it couldn't
     * process the directive. See if we can do it here.*/
    char *tmp2;
    //////////////////
    //return clause
    //////////////////
    if(strncmp(tmp, "return", 6) == 0)
    {
      tmp2 = tmp;
      int l = 6;
      while(is_whitespace(tmp2[l])) l++;
      if(l == 6 || tmp2[l] == '\0')
      {
	*res = 0;
	free(str); return 1;
      }
      
      for(i = 0; i < func->total_vars; i++)
      {
	int var_len = strlen(func->var_name[i]);
	if(strncmp(tmp2+l, func->var_name[i], var_len) == 0)
	{
	  if((tmp2[l+var_len] != '\n')
	     && (tmp2[l+var_len] != '\0')) break;
	  *res = func->var_value[i];
	  free(str); return 1;
	}
      }
      printf("Error: Variable %s is not defined in function %s()\n",
	     tmp2+l, func->var_name[i]);
      free(str); return 0;
    }
    char ERROR = 0;
    long result = 0;
    double result_f = 0;
    result_f = calc(tmp, &ERROR, &COMMENT, func);
    if(number_mode == NUMBER_INT)
      result = (int)result_f;
    if(COMMENT || ERROR) continue;
    //////////////////////////////////////////////
    //output the result
    //////////////////////////////////////////////
    output_res(output_format, result_f, result);
  }//end while
  /*
   * If we reached here, that means there was no 'return' clause.
   * Return '2' as a flag indicating this condition to the caller,
   * to suppress outputting function result.
   */
  free(str);
  return 2;
} /* process_custom_func() */
