/* 
 *    Programmed By: Mohammed Isam Mohammed [mohammed_isam1984@yahoo.com]
 *    Copyright 2014, 2015 (c)
 * 
 *    file: math.c
 *    This file is part of tcalc.
 *
 *    tcalc is free software: you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation, either version 3 of the License, or
 *    (at your option) any later version.
 *
 *    tcalc is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with tcalc.  If not, see <http://www.gnu.org/licenses/>.
 */    

#include <unistd.h>
#include "defs.h"

//extern char *strndup(const char *S, size_t SIZE);

/* function names for standard math functions */
char *func_name[] =
{
  "acos", "asin", "atan", "atan2", "cos", "sin", "tan",
  "cosh", "sinh", "tanh", "acosh", "asinh", "atanh",
  "exp", "log", "log10", "pow",
  "sqrt", "cbrt", "ceil", "floor", "fabs", "round",
  "trunc"
};
int total_funcs = 24;

/* function descriptions */
char *func_description[] =
{
  "acos(x)        Arc cosine of X\n",
  "asin(x)        Arc sine of X\n",
  "atan(x)        Arc tangent of X\n",
  "atan2(y, x)    Arc tangent of Y/X\n",
  "cos(x)         Cosine of X\n",
  "sin(x)         Sine of X\n",
  "tan(x)         Tangent of X\n",
  "cosh(x)        Hyperbolic cosine of X\n",
  "sinh(x)        Hyperbolic sine of X\n",
  "tanh(x)        Hyperbolic tangent of X\n",
  "acosh(x)       Hyperbolic arc cosine of X\n",
  "asinh(x)       Hyperbolic arc sine of X\n",
  "atanh(x)       Hyperbolic arc tangent of X\n",
  "exp(x)         Exponential function of X\n",
  "log(x)         Natural logarithm of X\n",
  "log10(x)       Base-ten logarithm of X\n",
  "pow(x, y)      X to the power Y\n",
  "sqrt(x)        Square root of X\n",
  "cbrt(x)        Cube root of X\n",
  "ceil(x)        Smallest integral value not less than X\n",
  "floor(x)       Largest integer not greater than X\n",
  "fabs(x)        Absolute value of X\n",
  "round(x)       Round X to nearest integral value, rounding halfway cases away\n"
  "               from zero\n",
  "trunc(x)       Round X to the integral value in floating-point format nearest\n"
  "               but not larger in magnitude\n"  
};

/*
 * Checks if *name is a reserved math function.
 */
int is_reserved_func(char *name, int len1)
{
  int j;
  for(j = 0; j < total_funcs; j++)
  {
    int len2 = strlen(func_name[j]);
    if(len1 != len2) continue;
    if(strncmp(func_name[j], name, len1) == 0)
    {
      printf("Error: %s() is a reserved function\n", func_name[j]);
      return 1;
    }
  }
  return 0;
}

/*
 * Prints function descriptions.
 */
void print_func_descriptions()
{
  int i;
  printf("Math functions:\n");
  for(i = 0; i < total_funcs; i++)
    printf("%s", func_description[i]);
}


/*
 * Process the passed function as indicated by its name in *func.
 */
double process_math_func(char *func, int ob_index, int cb_index,
			 char *error, struct custom_func_s *active_func)
{
  *error = 0;
  char *tmp = strndup(func, ob_index);
  if(!tmp)
  { 
    fprintf(stderr, "Error: Insufficient memory\n");
    *error = 1; return 0;
  }
  ob_index++;	//point to character after the opening brace
  char *tmp2 = 0, *tmp3;
  double res = -1;
  int i;
  //check math functions
  for(i = 0; i < total_funcs; i++)
  {
    if(strcmp(tmp, func_name[i]) == 0)
    {
      switch(i)
      {
	case 0:		//acos()
	  tmp2 = strndup(func+ob_index, cb_index-ob_index);
	  res = acos(atof(tmp2)); break;
	case 1:		//asin()
	  tmp2 = strndup(func+ob_index, cb_index-ob_index);
	  res = asin(atof(tmp2)); break;
	case 2:		//atan()
	  tmp2 = strndup(func+ob_index, cb_index-ob_index);
	  res = atan(atof(tmp2)); break;
	case 3:		//atan2()
	  if(!strchr(func+ob_index, ','))
	  {
	    printf("Error: missing argument: atan2(y, x)\n");
	    goto return_err;
	  }
	  i = strchr(func+ob_index, ',')-(func+ob_index);
	  tmp2 = strndup(func+ob_index, i);
	  cb_index = strchr(func+ob_index, ')')-(func+ob_index);
	  tmp3 = strndup(func+ob_index+i+1, cb_index-i);
	  res = atan2(atof(tmp2), atof(tmp3));
	  free(tmp3); break;
	case 4:		//cos()
	  tmp2 = strndup(func+ob_index, cb_index-ob_index);
	  res = cos(atof(tmp2)); break;
	case 5:		//sin()
	  tmp2 = strndup(func+ob_index, cb_index-ob_index);
	  res = sin(atof(tmp2)); break;
	case 6:		//tan()
	  tmp2 = strndup(func+ob_index, cb_index-ob_index);
	  res = tan(atof(tmp2)); break;
	case 7:		//cosh()
	  tmp2 = strndup(func+ob_index, cb_index-ob_index);
	  res = cosh(atof(tmp2)); break;
	case 8:		//sinh()
	  tmp2 = strndup(func+ob_index, cb_index-ob_index);
	  res = sinh(atof(tmp2)); break;
	case 9:		//tanh()
	  tmp2 = strndup(func+ob_index, cb_index-ob_index);
	  res = tanh(atof(tmp2)); break;
	case 10:	//acosh()
	  tmp2 = strndup(func+ob_index, cb_index-ob_index);
	  res = acosh(atof(tmp2)); break;
	case 11:	//asinh()
	  tmp2 = strndup(func+ob_index, cb_index-ob_index);
	  res = asinh(atof(tmp2)); break;
	case 12:	//atanh()
	  tmp2 = strndup(func+ob_index, cb_index-ob_index);
	  res = atanh(atof(tmp2)); break;
	case 13:	//exp()
	  tmp2 = strndup(func+ob_index, cb_index-ob_index);
	  res = exp(atof(tmp2)); break;
	case 14:	//log()
	  tmp2 = strndup(func+ob_index, cb_index-ob_index);
	  res = log(atof(tmp2)); break;
	case 15:	//log10()
	  tmp2 = strndup(func+ob_index, cb_index-ob_index);
	  res = log10(atof(tmp2)); break;
	case 16:	//pow()
	  if(!strchr(func+ob_index, ','))
	  {
	    printf("Error: missing argument: pow(x, y)\n");
	    *error = 1; return 0;
	  }
	  i = strchr(func+ob_index, ',')-(func+ob_index);
	  tmp2 = strndup(func+ob_index, i);
	  cb_index = strchr(func+ob_index, ')')-(func+ob_index);
	  tmp3 = strndup(func+ob_index+i+1, cb_index-i);
	  res = pow(atof(tmp2), atof(tmp3));
	  free(tmp3); break;
	case 17:	//sqrt()
	  tmp2 = strndup(func+ob_index, cb_index-ob_index);
	  res = sqrt(atof(tmp2)); break;
	case 18:	//cbrt()
	  tmp2 = strndup(func+ob_index, cb_index-ob_index);
	  res = cbrt(atof(tmp2)); break;
	case 19:	//ceil()
	  tmp2 = strndup(func+ob_index, cb_index-ob_index);
	  res = ceil(atof(tmp2)); break;
	case 20:	//floor()
	  tmp2 = strndup(func+ob_index, cb_index-ob_index);
	  res = floor(atof(tmp2)); break;
	case 21:	//fabs()
	  tmp2 = strndup(func+ob_index, cb_index-ob_index);
	  res = fabs(atof(tmp2)); break;
	case 22:	//round()
	  tmp2 = strndup(func+ob_index, cb_index-ob_index);
	  res = round(atof(tmp2)); break;
	case 23:	//trunc()
	  tmp2 = strndup(func+ob_index, cb_index-ob_index);
	  res = trunc(atof(tmp2)); 
	  //printf("tmp2 '%s', res '%f'\n", tmp2, res);
	  break;
      }//end switch
      if(tmp2) free(tmp2);
      break;
    }//end if
  }//end for
  if(i == total_funcs)
  {
    struct custom_func_s *f = &custom_func_head;
    while(f)
    {
      if(strcmp(tmp, f->custom_func_name) == 0)
      {
	tmp2 = strndup(func+ob_index, cb_index-ob_index);
	int h = process_custom_func(f, tmp2, &res);
	free(tmp2);
	/* means process_custom_func() encountered '{' or '}' */
	if(h == 2) goto return_err;
	if(!h)
	{
	    printf("Error: Function %s() returned error\n", tmp);
	    goto return_err;
	}
	break;
      }
      f = f->next;
    }
    if(f == NULL)
    {
      printf("Error: Unknown function: %s\n", tmp);
      goto return_err;
    }
  }
  free(tmp);
  return res;

return_err:
  free(tmp);
  *error = 1;
  return 0;
} /* process_math_func() */
